module CURIC::DCUI
  class DCUITool
    include Utils

    attr_accessor :attribute

    def initialize(attribute = nil)
      @attribute = attribute
      @model = Sketchup.active_model
      @selection = @model.selection
      @view = @model.active_view
    end

    def activate
      # get selected object
      @selected = selected_object

      @len_dimension = @selected.each_with_object({}) do |e, h|
        dimensions = compute_len_dimensions(e)[attribute['name'].to_sym]

        dimension = len_dimension_on_screen(e, dimensions)
        planes = len_planes(e, attribute['name'].to_sym)

        h[e] = [dimension, planes]
      end

      # ap @len_dimension
    end

    def reset
      @plane = nil
      @plane_points = nil
      @point = nil
      @vector = nil
    end

    def len_dimension_on_screen(e, dimensions)
      vector = current_entity_view(e)

      dimensions = dimensions.find_all do |d| 
        d[1].perpendicular?(vector)
      end
      dimensions_in = dimensions.find_all do |d|
        point_inside_screen?(midpoint(d[0]).offset(d[1], d[1].length / 2))
      end
      dimensions_see = dimensions_in.find_all do |d|
        point_can_see?(midpoint(d[0]).offset(d[1], d[1].length + 1.mm)) 
      end

      dimensions_see.sort_by! do |d|
        midpoint(d[0]).offset(d[1]).distance(@view.camera.eye)
      end

      return dimensions_see[0] if dimensions_see[0]
      return dimensions_in[0] if dimensions_in[0]

      dimensions[0]
    end

    def point_inside_screen?(point)
      point2d = @view.screen_coords(point)
      w = @view.vpwidth
      h = @view.vpheight
      screen = [
        [0, 0, 0],
        [w, 0, 0],
        [w, h, 0],
        [0, h, 0]
      ]

      Geom.point_in_polygon_2D(point2d, screen, false)
    end

    def point_can_see?(point)
      !@model.raytest([point, @view.camera.direction.reverse])
    end

    # Tính các dimensions của lenx, leny, lenz
    # Dimension tính từ center
    # Mỗi Lenx có 1 dimension và 4 hướng thể hiện
    def compute_len_dimensions(e)
      center = Utils.bound_center(e)
      lenx = instance_lenx(e)
      leny = instance_leny(e)
      lenz = instance_lenz(e)

      vecx = instance_vecx(e)
      vecx.length = lenx if lenx > 0
      vecy = instance_vecy(e)
      vecy.length = leny if leny > 0
      vecz = instance_vecz(e)
      vecz.length = lenz if lenz > 0

      line_x = [
        center.offset(vecx, lenx / 2),
        center.offset(vecx.reverse, lenx / 2)
      ]
      line_y = [
        center.offset(vecy, leny / 2),
        center.offset(vecy.reverse, leny / 2)
      ]
      line_z = [
        center.offset(vecz, lenz / 2),
        center.offset(vecz.reverse, lenz / 2)
      ]
      
      {
        lenx: [
          [line_x, vecy],
          [line_x, vecy.reverse],
          [line_x, vecz],
          [line_x, vecz.reverse]
        ],
        leny: [
          [line_y, vecx],
          [line_y, vecx.reverse],
          [line_y, vecz],
          [line_y, vecz.reverse]
        ],
        lenz: [
          [line_z, vecx],
          [line_z, vecx.reverse],
          [line_z, vecy],
          [line_z, vecy.reverse]
        ]
      }
    end

    def selected_object
      instances = @selection.grep(Sketchup::Group) + 
                  @selection.grep(Sketchup::ComponentInstance)
          
      instances
    end

    def deactivate(view)
      view.invalidate
    end

    def onMouseMove(*_args)
      @model.tools.pop_tool
    end

    def draw(view)
      return unless @len_dimension

      @len_dimension.each do |e, dim| 
        dimension, planes = dim
        planes.each { |plane| draw_plane(plane, view) }
        draw_dim(e, dimension, view) 
      end

      # draw_plane(view)
      # draw_direction(view)
    end
    
    def draw_dim(_e, dim, view, color = "blue")
      line, vec_offset = dim

      p0, p1 = line
      d = p0.distance(p1)
      text = "#{@attribute['formlabel']} = #{d.to_l}"

      # diagonal_line = bound_diagonal_line(e).map{ |pt| view.screen_coords(pt) }
      # d = diagonal_line[0].distance(diagonal_line[1])/10
      # p diagonal_line[0].distance(diagonal_line[1])
      d = view.pixels_to_model(100, midpoint(line))

      dim_line = line.map { |point| point.offset(vec_offset, vec_offset.length / 2 + d) }
      dim_extension_lines = line.map { |point| [point, point.offset(vec_offset, vec_offset.length / 2 + d)] }
      
      origin = midpoint(dim_line)
      
      # location text
      bb = Geom::BoundingBox.new
      return unless origin
      
      bb.add(origin)
      position = view.screen_coords(bb.max)
      
      draw_dim_line(view, dim_line, true, color)
      draw_dim_extension_lines(view, dim_extension_lines, color)
      draw_dim_text(position, text, view, color)
    end
    
    def draw_dim_extension_lines(view, lines, color = 'blue')
      lines.each { |line| draw_dim_line(view, line, false, color) }
    end
 
    def draw_dim_line(view, line, end_point = true, color = 'blue')
      line_2d = line.map { |pt| view.screen_coords(pt) }

      view.drawing_color = 'gray'
      view.line_stipple = '-'
      view.draw2d(GL_LINES, line_2d)

      view.line_stipple = ''
      # view.drawing_color = 'white'
      # view.line_width = 2
      # view.draw_line(line)
      view.line_width = 1
      view.drawing_color = color
      view.draw_line(line)

      view.draw_points(line, 7 * SCALE_FACTOR, 2, color) if end_point
    end

    def draw_dim_text(position, text, view, color = "blue", align = TextAlignCenter)
      color = Sketchup::Color.new(color)

      options = {
        font: "Arial",
        size: 12 * SCALE_FACTOR,
        # :bold => true,
        align: align,
        vertical_align: TextVerticalAlignBoundsTop,
        color: color
      }

      # border text
      bounds = view.text_bounds(position, text, options)
      width = bounds.width + 10
      height = bounds.height + 2
      # width, height = @textfont.compute_border(text, options[:size] + 2)
    
      # x1, y1 = bounds.upper_left.to_a
      # x2, y2 = bounds.lower_right.to_a
      # @points = [
      #   Geom::Point3d.new(x1, y1),
      #   Geom::Point3d.new(x1, y2),
      #   Geom::Point3d.new(x2, y2),
      #   Geom::Point3d.new(x2, y1),
      # ]

      width += 2
      
      rect = [
        Geom::Point3d.new(0, 0, 0),
        Geom::Point3d.new(width, 0, 0),
        Geom::Point3d.new(width, height, 0),
        Geom::Point3d.new(0, height, 0)
      ]
      rect_border = [
        Geom::Point3d.new(-1, -1, 0),
        Geom::Point3d.new(width + 1, -1, 0),
        Geom::Point3d.new(width + 1, height + 1, 0),
        Geom::Point3d.new(-1, height + 1, 0)
      ]

      tr = Geom::Transformation.new([0, 0, 0].vector_to(position))
      if options[:align] == TextAlignCenter
        point_move = Geom::Point3d.new(position.x - (width / 2), position.y, 0)
        vector = position.vector_to(point_move) 
        tr_move = Geom::Transformation.new(vector)
        tr *= tr_move
      elsif options[:align] == TextAlignLeft
        point_move = Geom::Point3d.new(position.x - 8, position.y, 0)
        vector = position.vector_to(point_move) 
        tr_move = Geom::Transformation.new(vector)
        tr *= tr_move
      end

      rect.map { |point| point.transform! tr }
      rect_border.map { |point| point.transform! tr }

      view.drawing_color = color
      view.draw2d(GL_QUADS, rect_border)

      view.drawing_color = "white"
      view.draw2d(GL_QUADS, rect)

      view.line_width = 2
      view.drawing_color = color
      view.line_stipple = ''
      # view.draw2d(GL_LINE_LOOP, rect)

      status = view.draw_text(position, text, options)
    end
    
    def draw_plane(plane, view)
      # d = view.pixels_to_model(5, midpoint(plane[0], plane[2]))
      # plane = @plane_points.map { |pt| pt.offset(@vector.reverse, d) }

      view.drawing_color = Sketchup::Color.new(255, 128, 128, 100)
      view.drawing_color = Sketchup::Color.new(0, 95, 163, 60)
      view.draw(GL_QUADS, plane)
      
      # Outline
      
      # 2d
      plane_2d = plane.map { |pt| view.screen_coords(pt) }
      view.line_width = 1
      view.line_stipple = '-' # Solid line
      view.drawing_color = 'gray'
      view.draw2d(GL_LINE_LOOP, plane_2d)
      
      view.line_width = 1
      view.line_stipple = '' # Solid line
      view.drawing_color = Sketchup::Color.new(255, 0, 0)
      view.draw(GL_LINE_LOOP, plane)

      # view.draw_points(plane, 10, 4, 'red')
    end

    def draw_direction(view)
      d = view.pixels_to_model(30, @point)
      line = [@point.offset(@vector.reverse, d * 3), @point.offset(@vector.reverse, d)]
      color = if @vector.parallel?(X_AXIS)
                'red'
              elsif @vector.parallel?(Y_AXIS)
                'lime'
              elsif @vector.parallel?(Z_AXIS)
                'blue'
              else
                [0, 95, 163]
              end
      # color = [0, 95, 163]
      view.drawing_color = color
      view.line_width = 1
      draw_arrow(line, view)
      view.line_width = 2
      view.draw_points(line[0], 10, 1, color)
    end

    def draw_arrow(line, view)
      # p line
      p1 = view.screen_coords(line[0])
      p2 = view.screen_coords(line[1])
      v = p2 - p1
      if v.valid?
        view.draw2d(GL_LINE_STRIP, [p1, p2])
        tt = Geom::Transformation.rotation p2, Geom::Vector3d.new(0, 0, 1), 160.degrees
        tn = Geom::Transformation.rotation p2, Geom::Vector3d.new(0, 0, 1), 90.degrees
        v = p2 - p1
        v.length = 30
        p3 = (p2 + v).transform(tt)
        p4 = (p2 + v).transform(tt.inverse)
        view.draw2d(GL_TRIANGLES, [p3, p4, p2])
      end
    end
  end
end
