module CURIC::DCUI
  # ##Settings
  # Thomas Thomassen :: Lib

  # Wrapper for +Sketchup.read_default+ and +Sketchup.write_default+.
  #
  # To initialize defaults:
  #   @settings = CURIC::Settings.new('FooBar')
  #   @settings.set_default( :foo, false )
  #   @settings.set_default( :bar, true )
  #   # Read
  #   read_value = @settings[ :foo ]
  #   # Write
  #   @settings[:bar] = false
  class Settings
    
    # Creates a new Settings instance that read and writes values to +section+
    # in SketchUp's preferences.
    #
    # @param [String] section
    def initialize(section)
      @section = section
      @cache = {}
    end     
    
    # Type casts read values based on the default value given where
    # +Sketchup.read_default+ would otherwise not do so.
    #
    # Custom type casts: +Length+, +Symbol+
    #
    # @param [String, Symbol] key
    # @param [mixed] default
    #
    # @return [mixed]
    def [](key, default = nil)
      if @cache.key?(key)
        x = @cache[key]
      else
        begin
          x = Sketchup.read_default(@section, key.to_s, default)
        rescue SyntaxError
          puts "#<Setting> Error reading setting! - Returning default value."
          puts "> #{@section.inspect} - #{key.to_s.inspect} (#{default.inspect})"
          x = default
        end
        x = x.to_l if default.is_a?(Length)
        x = x.intern if x.is_a?(String) && default.is_a?(Symbol)
        # x = true if x == 'Yes'
        # x = false if x == 'No'
        @cache[key] = x
        x
      end
    end     
    
    # Converts +Length+ to +Float+.
    #
    # Converts +Symbol+ to +String+.
    #
    # @param [String, Symbol] key
    # @param [mixed] value
    #
    # @return [mixed]
    def []=(key, value)
      @cache[key] = value
      value = value.to_f if value.is_a?(Length)
      value = value.to_s if value.is_a?(Symbol)
      Sketchup.write_default(@section, key.to_s, value)
      value
    end     
    
    # Type casts read values based on the default value given where
    # +Sketchup.read_default+ would otherwise not do so.
    #
    # Preferred over +settings[key, default]+ for setting defaults.
    #
    # Custom type casts: +Length+, +Symbol+
    #
    # @param [String, Symbol] key
    # @param [mixed] default
    #
    # @return [mixed]
    def set_default(key, default = nil)
      self.[](key, default)
    end     
    
    # Creates a sub-section.
    #
    # @deprecated Untested on OSX. Creates sub-keys under Windows.
    #
    # @param [String] sub_section
    #
    # @return [CURIC::Settings]
    def sub_section(sub_section)
      new("#{@section}\\#{sub_section}")
    end
  end

  module AttributeHelper
    module_function

    # Example:
    #   path = ['Curic', 'Section', 'Materials']
    #   create = true
    #   ad = AttributeHelper.dictionary(Sketchup.active_model, path, create)
    #   ad['Brick'] = m_brick_data
    #   ad.delete_key('Brick')
    # 
    # @return <Sketchup::AttributeDictionary>
    def dictionary(entity, path, create = false)
      begin
        path.each do |name|
          ads = entity.attribute_dictionaries
          create(entity, name) if (!ads || !ads[name]) && create

          entity = entity.attribute_dictionaries[name]
        end
      rescue StandardError => e
        entity = nil
      end

      entity
    end

    def create(entity, name)
      entity.set_attribute(name, 'temp', true)
      ad = entity.attribute_dictionaries[name]
      ad.delete_key('temp')

      true
    end
  end
  
  module Utils
    module_function

    # @return [Geom::Vector3d]
    def current_entity_view(e)
      tr = e.transformation
      vecx = tr.xaxis
      vecy = tr.yaxis
      vecz = tr.zaxis
        
      vecxa = vecx.reverse
      vecya = vecy.reverse
      vecza = vecz.reverse
      vecs = [vecx, vecy, vecz, vecxa, vecya, vecza]
      
      views = {}
      views[0] = RIGHT_VIEW
      views[1] = BACK_VIEW
      views[2] = TOP_VIEW
      views[3] = LEFT_VIEW
      views[4] = FRONT_VIEW
      views[5] = BOTTOM_VIEW
      
      v = e.model.active_view
      dicrection = v.camera.direction
      
      angles = {}
      
      (0..vecs.length - 1).each do |i|
        angle = vecs[i].angle_between(dicrection).radians
        angles[angle] = i
      end

      vecs[angles[angles.keys.min]]
    end
    
    def get_current_bound_view(bound, tr)
      point0 = bound.corner(0).transform(tr)

      point1 = bound.corner(1).transform(tr)
      point2 = bound.corner(2).transform(tr)
      point4 = bound.corner(4).transform(tr)

      vecx = point0.vector_to(point1)
      vecy = point0.vector_to(point2)
      vecz = point0.vector_to(point4)
        
      vecxa = vecx.reverse
      vecya = vecy.reverse
      vecza = vecz.reverse
      vecs = [vecx, vecy, vecz, vecxa, vecya, vecza]
      
      views = {}
      views[0] = RIGHT_VIEW
      views[1] = BACK_VIEW
      views[2] = TOP_VIEW
      views[3] = LEFT_VIEW
      views[4] = FRONT_VIEW
      views[5] = BOTTOM_VIEW
      
      v = Sketchup.active_model.active_view
      c = v.camera
      up = c.up
      target = c.target
      eye = c.eye
      dicrection = eye - target
      angles = {}
      (0..vecs.length - 1).each do |i|
        angle = (vecs[i].angle_between dicrection).radians
        angles[angle] = i
      end

      views[angles[angles.keys.min]]
    end

    # points bound in 2d
    def get_polygon_bounds(definition, view)
      bb = definition.bounds
      points = get_plane(definition, view)
      plane = Geom.fit_plane_to_points(points)
      vector = Geom::Vector3d.new(plane[0], plane[1], plane[2])
      polygon = nil

      polygon = if vector.parallel?(X_AXIS)
                  points.map { |point| [point.y, point.z, point.x] }
                elsif vector.parallel?(Y_AXIS)
                  points.map { |point| [point.x, point.z, point.y] }
                else
                  points
                end

      # transform to origin
      # các component có axes ko tại bounds.min cần chuyển points => origin
      # find upper_left
      upper_left = polygon[0]
      polygon.each do |point| 
        x = point.x
        y = point.y
        upper_left = point if upper_left.x >= x && upper_left.y >= y
      end
      tr = Geom::Transformation.new(upper_left)
      polygon.each { |point| point.transform! tr.inverse }
      polygon
    end

    # plane view
    def get_plane(definition, view)
      bb = definition.bounds
      bound_direction_plane(bb, view)
    end

    def bound_direction_plane(bounds, view)
      plane = nil

      if bounds && bounds.width != 0 && bounds.height != 0 && bounds.depth != 0
        tr_entity = Geom::Transformation.new
        pts = bound_points(bounds).map { |pt| pt.transform(tr_entity) }
        if pts.length == 8
          if view == TOP_VIEW
            plane = [pts[4], pts[6], pts[7], pts[5]]
          elsif view == BOTTOM_VIEW 
            plane = [pts[1], pts[3], pts[2], pts[0]]
          elsif view == FRONT_VIEW 
            plane = [pts[0], pts[4], pts[5], pts[1]]
          elsif view == BACK_VIEW
            plane = [pts[3], pts[2], pts[6], pts[7]]
          elsif view == LEFT_VIEW 
            plane = [pts[2], pts[0], pts[4], pts[6]]
          elsif view == RIGHT_VIEW
            plane = [pts[1], pts[3], pts[7], pts[5]]
          end
        end
      end

      plane
    end
    
    def len_planes(e, len)
      bounds = entity_bound(e)
      pts = bound_points(bounds).map { |pt| pt.transform(e.transformation) }
     
      top = [pts[4], pts[6], pts[7], pts[5]]
      bottom = [pts[1], pts[3], pts[2], pts[0]]
      front = [pts[0], pts[4], pts[5], pts[1]]
      back = [pts[3], pts[2], pts[6], pts[7]]
      left = [pts[2], pts[0], pts[4], pts[6]]
      right = [pts[1], pts[3], pts[7], pts[5]]

      case len
      when :lenx
        [left, right]
      when :leny
        [front, back]
      else
        [top, bottom]
      end
		end
		 
		def bound_diagonal_line(e)
      bounds = entity_bound(e)
      pts = bound_points(bounds).map { |pt| pt.transform(e.transformation) }
		 
			[pts[0], pts[7]]
		end
		 

    
    def entity_bound(e)
      e.definition.bounds
    end

    def entity_bound_points(e, tr = Geom::Transformation.new)
      b = entity_bound(e)
      bound_points(b, tr)
    end

    # BoundingBox Points
    def bound_points(bound, tr = Geom::Transformation.new)
      # if bound.width == 0
      #   [0,2,6,4].map { |i| bound.corner(i) }
      # elsif bound.height == 0
      #   [0,1,5,4].map { |i| bound.corner(i) }
      # elsif bound.depth == 0
      #   [0,1,3,2].map { |i| bound.corner(i) }
      # else
      (0..7).map { |i| bound.corner(i).transform(tr) }
      # end
    end

    def bound_center(e)
      entity_bound(e).center.transform(e.transformation)
    end
    
    def instance_lenx(e)
      b_pts = entity_bound_points(e, e.transformation)
      b_pts[0].distance(b_pts[1])
    end

    def instance_leny(e)
      b_pts = entity_bound_points(e, e.transformation)
      b_pts[0].distance(b_pts[2])
    end

    def instance_lenz(e)
      b_pts = entity_bound_points(e, e.transformation)
      b_pts[0].distance(b_pts[4])
    end

    def instance_vecx(e)
      e.transformation.xaxis
    end

    def instance_vecy(e)
      e.transformation.yaxis
    end

    def instance_vecz(e)
      e.transformation.zaxis
    end
    
    def midpoint(segment, point = nil)
      segment = [segment, point] if point
      Geom.linear_combination(0.5, segment.first, 0.5, segment.last) 
    end
  end

  
# https://github.com/munshkr/easing-ruby
# https://easings.net/vi

  module Easing
    extend self

    def linear_tween(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      c * t / d + b
    end

    def ease_in_quad(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return c*(t/=d)*t + b;
    end

    def ease_out_quad(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return -c *(t/=d)*(t-2) + b;
    end

    def ease_in_out_quad(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      t /= d / 2
      return c / 2*t*t + b if (t < 1)
      t -= 1
      return -c/2 * (t*(t-2) - 1) + b
    end

    def ease_in_cubic(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return c*(t/=d)*t*t + b
    end

    def ease_out_cubic(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return c*((t=t/d-1)*t*t + 1) + b
    end

    def ease_in_out_cubic(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return c/2*t*t*t + b if ((t/=d/2) < 1)
      return c/2*((t-=2)*t*t + 2) + b
    end

    def ease_in_quart(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return c*(t/=d)*t*t*t + b
    end

    def ease_out_quart(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return -c * ((t=t/d-1)*t*t*t - 1) + b
    end

    def ease_in_out_quart(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return c/2*t*t*t*t + b if ((t/=d/2) < 1)
      return -c/2 * ((t-=2)*t*t*t - 2) + b
    end

    def ease_in_quint(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return c*(t/=d)*t*t*t*t + b
    end

    def ease_out_quint(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return c*((t=t/d-1)*t*t*t*t + 1) + b
    end

    def ease_in_out_quint(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return c/2*t*t*t*t*t + b if ((t/=d/2) < 1)
      return c/2*((t-=2)*t*t*t*t + 2) + b
    end

    def ease_in_sine(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return -c * Math.cos(t/d * (Math::PI/2)) + c + b
    end

    def ease_out_sine(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return c * Math.sin(t/d * (Math::PI/2)) + b
    end

    def ease_in_out_sine(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return -c/2 * (Math.cos(Math::PI*t/d) - 1) + b
    end

    def ease_in_expo(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return (t==0) ? b : c * (2 ** (10 * (t/d - 1))) + b
    end

    def ease_out_expo(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return (t==d) ? b+c : c * (-2**(-10 * t/d) + 1) + b
    end

    def ease_in_out_expo(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return b if t == 0
      return b + c if t == d
      return (c/2) * 2**(10 * (t-1)) + b if ((t /= d/2) < 1)
      return (c/2) * (-2**(-10 * t-=1) + 2) + b
    end

    def ease_in_circ(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return -c * (Math.sqrt(1 - (t/=d)*t) - 1) + b
    end

    def ease_out_circ(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return c * Math.sqrt(1 - (t=t/d-1)*t) + b
    end

    def ease_in_out_circ(t, b, c, d)
      t = t.to_f; b = b.to_f; c = c.to_f; d = d.to_f

      return -c/2 * (Math.sqrt(1 - t*t) - 1) + b if ((t/=d/2) < 1)
      return c/2 * (Math.sqrt(1 - (t-=2)*t) + 1) + b
    end
  end

  # @animation = ResizeDialog.new(@dialog)
  # Sketchup.active_model.active_view.animation = @animation

  class ResizeDialog
    attr_accessor(:stop_animation, :width, :height)

    def initialize(dialog)
      @dialog = dialog

      @width = 500
      @height = 1000
      @new_width = 1000
      
      @types = [
        :ease_in_quad,
        :ease_out_quad,
        :ease_in_out_quad,
        :ease_in_cubic,
        :ease_out_cubic,
        :ease_in_out_cubic,
        :ease_in_quart,
        :ease_out_quart,
        :ease_in_out_quart,
        :ease_in_quint,
        :ease_out_quint,
        :ease_in_out_quint,
        :ease_in_sine,
        :ease_out_sine,
        :ease_in_out_sine,
        :ease_in_expo,
        :ease_out_expo,
        :ease_in_out_expo,
        :ease_in_circ,
        :ease_out_circ,
        :ease_in_out_circ
      ]

      @stop_animation = false
      @frame = 0
    end

    def onDone(&block)
      @ondone = block
    end

    def new_width=(w)
      @new_width = w
      if @new_width > @width
        @resize = :open
        @start_width = @width
        @end_width = @new_width
      else
        @resize = :close
        @start_width = @new_width
        @end_width = @width
      end
    end

    def nextFrame(view)
      return false if @stop_animation

      continue = if @resize == :open
                  @width < @new_width
                else
                  @width > @new_width
                end
      # p ['continue', continue, [@resize, @width, @new_width]]

      if continue
        begin
          width = Easing.ease_in_expo(@frame, @start_width, @end_width, 12)
          # p [@resize, @width, @start_width, @end_width]
          @width = if @resize == :open
            width > @new_width ? @new_width : width
          else
            w = @start_width + (@end_width - width)
            w < @new_width ? @new_width : w
          end

          @dialog.set_size(@width, @height)

          # @dialog.execute_script()
          # @dialog.set_position(@width, -500)
        rescue => exception
          continue = false
        end

        view.show_frame
      end
      
      if !continue
        p "resize done: #{@width}"
        @ondone.call if @ondone
      end

      @stop_animation = true unless continue

      @frame += 1
      
      return continue
    end
  end
end
